import sys,time,os, glob
from mpi4py import MPI
from seisgo import noise, utils

if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")

# absolute path parameters
rootpath  = "test_data"                                                     # root path for this data processing
CCFDIR    = os.path.join(rootpath,'CCF_test')                               # dir to store CC data
DATADIR   = os.path.join(rootpath,'raw_data')                               # dir where noise data is located
locations = os.path.join(rootpath,'station.txt')                            # station info including network,station,channel,latitude,longitude,elevation: only needed when input_fmt is not asdf

# some control parameters
freq_norm   = 'rma'                                                         # 'no' for no whitening, or 'rma' for running-mean average, 'phase' for sign-bit normalization in freq domain
time_norm   = 'no'                                                          # 'no' for no normalization, or 'rma', 'one_bit' for normalization in time domain
cc_method   = 'xcorr'                                                       # 'xcorr' for pure cross correlation, 'deconv' for deconvolution; FOR "COHERENCY" PLEASE set freq_norm to "rma" and time_norm to "no"
acorr_only  = False                                                         # only perform auto-correlation
xcorr_only  = True                                                          # only perform cross-correlation or not
correct_orientation = True                                                  # If Ture, correct orientations for horizontal channels and convert 1/2 to N/E channels
rotate_raw = True                                                          # If Ture, rotate to RTZ coordinate. correct_orientation will be automatically set to True
max_time_diff = 60                                                               # (unit: datapoints) For padding horizontal traces to match sizes when doing orient correction. If None, default is 10 data points.
exclude_chan = []                                                           # Added by Xiaotao Yang. Channels in this list will be skipped.

channel_pairs = ['ZZ','TT'] ### !!! IMPORTANT !!! Less than 3 pairs are strongly recommended for efficiency. Set to 'None' will do all cross-correlations

verbose = True
output_structure="source"                                                   # How the output files will be named. Acceptable formats are: "raw", "source", "station-pair", "station-component-pair", 
# pre-processing parameters
cc_len    = 3600*2                                                            # basic unit of data length for fft (sec)
step      = 3600*0.5                                                             # (1-overlapping) between each cc_len (sec). The smaller the more chuncks
smooth_N  = 150                                                              # Important. This value should be at least the max period of interet. Half moving window length for time/freq domain normalization if selected (points)

# cross-correlation parameters
maxlag         = 1500                                                        # lags of cross-correlation to save (sec)
substack       = False                                                      # sub-stack daily cross-correlation or not
substack_len   = cc_len                                                  # how long to stack over (for monitoring purpose): need to be multiples of cc_len
smoothspect_N  = 20                                                         # moving window length to smooth spectrum amplitude (points)

# criteria for data selection
max_over_std = 10                                                           # threahold to remove window of tremor signals: set it to 10*9 if prefer not to remove them
max_kurtosis = 10                                                           # max kurtosis allowed, TO BE ADDED!

# load useful download info if start from ASDF
dfile = os.path.join(DATADIR,'download_info.txt')
down_info = eval(open(dfile).read())
freqmin   = down_info['freqmin']
freqmax   = down_info['freqmax']
##################################################
# we expect no parameters need to be changed below
#--------MPI---------
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if rank == 0:
    # make a dictionary to store all variables: also for later cc
    fc_para={'cc_len':cc_len,'step':step,
            'freqmin':freqmin,'freqmax':freqmax,'freq_norm':freq_norm,'time_norm':time_norm,
            'cc_method':cc_method,'smooth_N':smooth_N,'substack':substack,'substack_len':substack_len,
            'smoothspect_N':smoothspect_N,'maxlag':maxlag,'max_over_std':max_over_std,
            'max_kurtosis':max_kurtosis,'channel_correction':correct_orientation,'rotate_raw':rotate_raw,'channel_pairs':channel_pairs}
    # save fft metadata for future reference
    fc_metadata  = os.path.join(CCFDIR,'fft_cc_data.txt')
    if not os.path.isdir(CCFDIR):os.makedirs(CCFDIR)
    # save metadata
    fout = open(fc_metadata,'w')
    fout.write(str(fc_para));fout.close()

    # set variables to broadcast
    tdir = sorted(glob.glob(os.path.join(DATADIR,'*.h5')))

    nchunk = len(tdir)
    splits  = nchunk

    if nchunk==0: raise IOError('Abort! no available seismic files for FFT')

else:
    splits,tdir = [None for _ in range(2)]

# broadcast the variables
splits = comm.bcast(splits,root=0)
tdir  = comm.bcast(tdir,root=0)
#loop through all data files.
for ick in range(rank,splits,size):
    sfile=tdir[ick]
    t10=time.time()
    #call the correlation wrapper.
    ndata,total_t=noise.do_correlation(sfile,cc_len,step,maxlag,channel_pairs,correct_orientation=correct_orientation,rotate_raw=rotate_raw,cc_method=cc_method,
                         acorr_only=acorr_only,xcorr_only=xcorr_only,substack=substack,
                         smoothspect_N=smoothspect_N,substack_len=substack_len,
                         maxstd=max_over_std,freqmin=freqmin,freqmax=freqmax,
                         time_norm=time_norm,freq_norm=freq_norm,smooth_N=smooth_N,max_time_diff=max_time_diff,
                         exclude_chan=exclude_chan,outdir=CCFDIR,v=verbose,output_structure=output_structure)

    t11 = time.time()
    print('it takes %6.5fs to process the chunk of %s' % (t11-t10,sfile.split('/')[-1]))
    if rotate_raw:
        print('it takes %6.5fs to assemble raw source data' % (total_t[1]))
        print('it takes %6.5fs to assemble raw receiver data' % (total_t[5]))
        print('it takes %6.5fs to do rotation' % (total_t[2]))
        print('it takes %6.5fs to prepare data for FFT' % (total_t[3]))
        print('it takes %6.5fs to compute FFT' % (total_t[4]))

#comm.barrier()

# merge all path_array and output
#if rank == 0:
#    sys.exit()



#!/usr/bin/env python
# coding: utf-8

# # Tilt and Compliance Corrections for OBS Data: Continuous
# ### Xiaotao Yang @ Harvard University
# This notebook contains examples of compliance corrections using local data on the disk.
# The functions for tilt and compliance corrections are in module seisgo.obsmaster.

# ## Step 0. Load needed packages.
# Some functions are imported from the utils.py and the obsmaster.py.

#import needed packages.
from seisgo import utils
from seisgo import obsmaster as obs
import sys
import time
import scipy
import obspy
import pyasdf
import datetime
import os, glob
import numpy as np
import pandas as pd
# import matplotlib.pyplot  as plt
# from obspy import UTCDateTime
from obspy.core import Stream, Trace

"""
1. Set global data path parameters.
"""
rawdatadir = '../data/raw'
if not os.path.isdir(rawdatadir): os.mkdir(rawdatadir)
#directory to save the data after TC removal
tcdatadir = '../data/tcremoval'
if not os.path.isdir(tcdatadir): os.mkdir(tcdatadir)
####################################
#### Optional clean-up block ####
####################################
cleantartgetdir=True #change to False or remove/comment this block if needed.
dfiles0 = glob.glob(os.path.join(tcdatadir,'*.h5'))
if cleantartgetdir and len(dfiles0)>0:
    print('Cleaning up TC removal directory before running ...')
    for df0 in dfiles0:os.remove(df0)
####################################
##### End of clean-up block #####
####################################

####################################
#### Optional downloading block ####
####################################
downloadexample=False #change to False or remove/comment this block if needed.
if downloadexample:
    print('Cleaning up raw data directory before downloading ...')
    dfiles1 = glob.glob(os.path.join(rawdatadir,'*.h5'))
    for df1 in dfiles1:os.remove(df1)
    os.system('python seisgo_download_obsdata.py')
####################################
##### End of downloading block #####
####################################

"""
2. Tilt and compliance removal parameters
"""
window=3600
overlap=0.2
taper=0.08
qc_freq=[0.004, 1]
plot_correction=True
normalizecorrectionplot=True
tc_subset=['ZP-H']
#assemble all parameters into a dictionary.
tcpara={'window':window,'overlap':overlap,'taper':taper,'qc_freq':qc_freq,
        'tc_subset':tc_subset}
print(tcpara)
"""
3. Read local data and do correction
We use the wrapper function for tilt and compliance corrections.

Steps:
a. read in file list
b. loop through all files
c. loop through all stations
    c-1. read waveform tags and list
    c-2. read station info if available. skip land stations or stations with only vertical.
    c-3. assemble all four components for OBS stations
    c-4. do correction work flow and plot the result if applicable
    c-5. save auxiliary data, e.g., tilt direction and angle, and TC removal parameters.
    c-6. save to original file name in different directory
"""
dfiles = glob.glob(os.path.join(rawdatadir,'*.h5'))
nfiles = len(dfiles)
splits  = nfiles
if nfiles==0:
    raise IOError('Abort! no available seismic files in '+rawdatadir)

t0=time.time()
for ifile in range(nfiles):
    df=dfiles[ifile]
    print('Working on: '+df+' ... ['+str(ifile+1)+'/'+str(nfiles)+']')
    dfbase=os.path.split(df)[-1]
    df_tc=os.path.join(tcdatadir,dfbase)

    ds=pyasdf.ASDFDataSet(df,mpi=False,mode='r')
    netstalist = ds.waveforms.list()
    nsta = len(netstalist)

    tilt=[]
    sta_processed=[]
    for ista in netstalist:
        print('  station: '+ista)
        """
        Get the four-component data
        """
        try:
            inv = ds.waveforms[ista]['StationXML']
        except Exception as e:
            print('  No stationxml for %s in file %s'%(ista,df))
            inv = None

        all_tags = ds.waveforms[ista].get_waveform_tags()
        print(all_tags)
        if len(all_tags)!=4:
            print("  Wrong number of components. Has to be four (4) channels! Skip!")
            continue

        tr1=None
        tr2=None
        trZ=None
        trP=None
        #assign components by waveform tags.
        #This step may fail if the tags don't reflect the real channel information
        newtags=['-','-','-','-']
        for tg in all_tags:
            tr_temp = ds.waveforms[ista][tg][0]
            chan = tr_temp.stats.channel
            if chan[-1].lower() == 'h':trP=tr_temp;newtags[3]=tg
            elif chan[-1].lower() == '1' or chan[-1].lower() == 'e':tr1=tr_temp;newtags[0]=tg
            elif chan[-1].lower() == '2' or chan[-1].lower() == 'n':tr2=tr_temp;newtags[1]=tg
            elif chan[-1].lower() == 'z':trZ=tr_temp;newtags[2]=tg

        #sanity check.
        for tr in [tr1, tr2, trZ, trP]:
            if not isinstance(tr, Trace):
                print(str(tr)+" is not a Trace object")

        """
        Call correction wrapper
        """
        spectra,transfunc,correct=obs.TCremoval_wrapper(
            tr1,tr2,trZ,trP,window=window,overlap=overlap,merge_taper=taper,
            qc_freq=qc_freq,qc_spectra=True,fig_spectra=False,
            save_spectrafig=False,fig_transfunc=False,correctlist=tc_subset)
        tilt.append(spectra['rotation'].tilt)
        sta_processed.append(ista)
        if plot_correction:
            obs.plotcorrection(trZ,correct,normalize=normalizecorrectionplot,freq=[0.005,0.1],
                               size=(12,3),save=True,form='png')

        """
        Save to ASDF file.
        """
        trZtc,tgtemp=obs.correctdict2stream(trZ,correct,tc_subset)
        print('  saving to: '+df_tc)
        utils.save2asdf(df_tc,Stream(traces=[tr1,tr2,trZtc[0],trP]),newtags,sta_inv=inv)

    #save auxiliary data to file.
    print('  saving auxiliary data to: '+df_tc)
    tcpara_temp=tcpara
    tcpara_temp['tilt_stations']=sta_processed
    utils.save2asdf(df_tc,np.array(tilt),None,group='auxiliary',para={'data_type':'tcremoval',
                                                'data_path':'tiltdir',
                                                'parameters':tcpara_temp})

tend=time.time() - t0
print('Finished all files in '+str(tend)+' seconds, or '+str(tend/3600)+' hours')

# In[ ]:



import os,glob,copy,obspy,scipy,time,pycwt,pyasdf,datetime
import numpy as np
import pandas as pd
from scipy.signal import hilbert
from obspy.signal.util import _npts2nfft
from obspy.signal.invsim import cosine_taper
from obspy.signal.regression import linear_regression
from obspy.core.util.base import _get_function_from_entry_point
from obspy.core.inventory import Inventory, Network, Station, Channel, Site
from scipy.fftpack import fft,ifft,next_fast_len
from obspy.signal.filter import bandpass,lowpass
import matplotlib.pyplot as plt
"""
This is a planned module, to be developed.
"""
################################################################
################ DISPERSION EXTRACTION FUNCTIONS ###############
################################################################
def get_dispersion_waveforms_cwt(d, dt,fmin,fmax,dj=1/12, s0=-1, J=-1, wvn='morlet'):
    """
    Produce dispersion wavefroms with continuous wavelet tranform.

    ===parameters===
    d: 1-d array data.
    df: time interval.
    fmin, fmax: frequency range.
    dj=1/12, s0=-1, J=-1, wvn='morlet': pycwt.cwt parameters.

    ==returns===
    dout, fout: narrowband-filtered waveforms and the frequency vector.
    """
    ds_cwt, sj, f, coi, _, _ = pycwt.cwt(d, dt, dj, s0, J, wvn)
    f_ind = np.where((f >= fmin) & (f <= fmax))[0]
    dout=[]
    fout=[]
    for ii in range(len(f_ind)):
        if ii>0 and ii<len(f_ind)-1: f_ind_temp=f_ind[ii-1:ii+1]
        elif ii==len(f_ind)-1: f_ind_temp=f_ind[ii-1:ii]
        elif ii==0:f_ind_temp=f_ind[ii:ii+1]
        fout.append(np.mean(f[f_ind_temp]))
        rds_cwt=np.real(pycwt.icwt(ds_cwt[f_ind_temp], sj[f_ind_temp], dt, dj, wvn))
        ds_win=np.power(rds_cwt,2)
        dout.append(ds_win/np.max(ds_win))
    return np.flip(np.array(dout),axis=0), np.flip(fout)

def narrowband_waveforms(d, dt,pmin,pmax,dp=1,pscale='ln',extend=10):
    """
    Produce dispersion wavefroms with narrowband filters.

    ===parameters===
    d: 1-d array data.
    dt: sampling interval.
    pmin, pmax: period range.
    dp: period increment in seconds. default 1 s.
    pscale: period scales. "ln" for linear [default]. "nln" for non-linear scale.
    extend: extend individual period value to form a band range. default: 5 scale steps.

    ==returns===
    dout, pout: narrowband-filtered waveforms and the period vector.
    """
    period=np.array([pmin - extend*dp,pmax + extend*dp])
    if period[0] < 2*dt: period[0]=2.01*dt

    if pscale=="ln":
        # f_all=np.arange(fmin-extend*df,fmax+extend*df,df)
        ptest=np.arange(period.min(),period.max(),dp)
    elif pscale=="nln":
        ptest=2 ** np.arange(np.log2(0.1*period.min()),
                    np.log2(2*period.max()),dp)
    f_all=np.flip(1/ptest)
    fout_temp=[]
    dout_temp=[]
    din=d.copy()

    for ii in range(len(f_all)-extend):
        if f_all[ii]>=1/(2*dt) or f_all[ii+extend]>=1/(2*dt): continue
        ds_win=bandpass(din,f_all[ii],f_all[ii+extend],1/dt,corners=4, zerophase=True)
        dout_temp.append(ds_win/np.max(np.abs(ds_win)))
        fout_temp.append(np.mean([f_all[ii],f_all[ii+extend]])) #center frequency
    fout_temp=np.array(fout_temp)
    f_ind=np.where((fout_temp>=1/pmax) & (fout_temp<=1/pmin))[0]
    fout=fout_temp[f_ind]
    dout_temp=np.array(dout_temp)
    dout = dout_temp[f_ind]
    pout = 1/fout
    return dout, pout
##
def get_dispersion_image(g,t,d,pmin,pmax,vmin,vmax,dp=1,dv=0.1,window=1,pscale='ln',pband_extend=5,
                        verbose=False,min_trace=5,min_wavelength=1.5,energy_type='power_sum',
                        plot=False,figsize=None,cmap='jet',clim=[0,1]):
    """
    Uses phase-shift method. Park et al. (1998): http://www.masw.com/files/DispersionImaingScheme-1.pdf

    =====PARAMETERS====
    g: waveform gather for all distances (traces). It should be a numpy array.
    t: time vector.
    d: distance vector corresponding to the waveforms in `g`
    pmin: minimum period.
    pmax: maximum period.
    vmin: minimum phase velocity to search.
    vmax: maximum phase velocity to search.
    dp: period increment. default is 1.
    dv: velocity increment in searching. default is 0.1
    window: number of wavelength when slicing the time segments in computing summed energy. default is 1.
            Window can be a two-element array [min,max], when the window size will be interpolated between
            the minimum and the maximum.
    pscale: period vector scale in applying narrowband filters. default is 'ln' for linear scale.
    pband_extend: number of period increments to extend in filtering. defult is 5.
    verbose: verbose mode. default False.
    min_trace: minimum trace to consider. default 5.
    min_wavelength: minimum wavelength to satisfy far-field. default 1.5.
    energy_type: method to compute maximum energy, 'envelope' or 'power_sum'. Default is 'power_sum'
    plot: plot dispersion image or not. Default is False.
    figsize: specify figsize. Decides automatically if not specified.
    cmap: colormap. Default is 'jet'
    clim: color value limit. Default is [0,1]

    =====RETURNS====
    dout: dispersion information showing the normalized energy for each velocity value for each frequency.
    vout: velocity vector used in searching.
    pout: period vector.
    """
    #validate options.
    energy_type_list=['power_sum','envelope']
    if energy_type.lower() not in energy_type_list:
        raise ValueError(energy_type+" is not a recoganized energy type. Use one of "+energy_type_list)
    if len(np.array(window).shape) < 1:
        window=[window,window]

    dt=np.abs(t[1]-t[0])
    if t[0]<-1.0*dt and t[-1]> dt: #two sides.
        side='a'
        zero_idx=int((len(t)-1)/2)
        if figsize is None:
            figsize=(8,3)
    elif t[0]<-1.0*dt:
        side='n'
        zero_idx=len(t)-1
        if figsize is None:
            figsize=(5,4)
    elif t[-1]>dt:
        side='p'
        zero_idx=0
        if figsize is None:
            figsize=(5,4)

    if verbose: print('working on side: '+side)
    dfiltered_all=[]
    dist_final=[]
    for k in range(g.shape[0]):
        dtemp,pout=narrowband_waveforms(g[k]/np.max(np.abs(g[k])),dt,pmin,
                                        pmax,dp=dp,pscale=pscale,extend=pband_extend)
        dfiltered_all.append(dtemp)
    dfiltered_all=np.array(dfiltered_all)
    vout=np.arange(vmin,vmax+0.5*dv,dv)
    dout_n_all=[]
    dout_p_all=[]
    window_vector=np.linspace(window[1],window[0],len(pout))
    for k in range(len(pout)):
        win_length=window_vector[k]*pout[k]
        win_len_samples=int(win_length/dt)+1
        dout_n=[]
        dout_p=[]

        d_in=dfiltered_all[:,k,:]
        for i,v in enumerate(vout):
            #subset by distance
            mindist=min_wavelength*v*pout[k] #at least 1.5 wavelength.
            dist_idx=np.where((d >= mindist))[0]
            if len(dist_idx) >min_trace:
                if side=='a' or side=='n':
                    dvec=[]
                    for j in dist_idx: #distance, loop through traces
                        tmin=d[j]/v
                        tmin_idx=zero_idx - int(tmin/dt)
                        dsec=d_in[j][tmin_idx - win_len_samples : tmin_idx]
                        if not any(np.isnan(dsec)) and len(dsec)==win_len_samples:
                            dvec.append(dsec)
                    if energy_type.lower() == 'power_sum':
                        peak_energy=np.sum(np.power(np.mean(dvec,axis=1),2))
                    elif energy_type.lower() == 'envelope':
                        peak_energy=np.max(np.abs(hilbert(np.mean(dvec,axis=1))))
                    dout_n.append(peak_energy)

                if side=='a' or side=='p':
                    dvec=[]
                    for j in dist_idx: #distance, loop through traces
                        tmin=d[j]/v
                        tmin_idx=zero_idx + int(tmin/dt)
                        dsec=d_in[j][tmin_idx : tmin_idx + win_len_samples]
                        if not any(np.isnan(dsec)) and len(dsec)==win_len_samples:
                            dvec.append(dsec)
                    #
                    if energy_type.lower() == 'power_sum':
                        peak_energy=np.sum(np.power(np.mean(dvec,axis=1),2))
                    elif energy_type.lower() == 'envelope':
                        peak_energy=np.max(np.abs(hilbert(np.mean(dvec,axis=1))))
                    dout_p.append(peak_energy)
            else:
                if side=='a' or side=='n':
                    dout_n.append(np.nan)

                if side=='a' or side=='p':
                    dout_p.append(np.nan)

        if side=='a' or side=='n':
            dout_n /= np.nanmax(dout_n)
            dout_n_all.append(dout_n)
        if side=='a' or side=='p':
            dout_p /= np.nanmax(dout_p)
            dout_p_all.append(dout_p)
    # plot or not
    if plot:
        plt.figure(figsize=figsize)
        if side == 'a':
            plt.subplot(1,2,1)
            plt.imshow(np.flip(np.array(dout_n_all).T),cmap=cmap,extent=[pout[-1],pout[0],vout[0],vout[-1]],aspect='auto')
            plt.ylabel('velocity (km/s)',fontsize=12)
            plt.xlabel('period (s)',fontsize=12)
            plt.xticks(np.linspace(pmin,pmax,5),fontsize=12)
            plt.yticks(np.linspace(vmin,vmax,5),fontsize=12)
            plt.clim(clim)
            ax=plt.colorbar()
            ax.set_label('normalized energy (%s)'%(energy_type.replace('_',' ')))
            plt.title('negative lag: '+energy_type.replace('_',' '),fontsize=13)

            plt.subplot(1,2,2)
            plt.imshow(np.flip(np.array(dout_p_all).T),cmap=cmap,extent=[pout[-1],pout[0],vout[0],vout[-1]],aspect='auto')
            plt.ylabel('velocity (km/s)',fontsize=12)
            plt.xlabel('period (s)',fontsize=12)
            plt.xticks(np.linspace(pmin,pmax,5),fontsize=12)
            plt.yticks(np.linspace(vmin,vmax,5),fontsize=12)
            plt.clim(clim)
            ax=plt.colorbar()
            ax.set_label('normalized energy (%s)'%(energy_type.replace('_',' ')))
            plt.title('positive lag: '+energy_type.replace('_',' '),fontsize=13)

            plt.tight_layout()
        elif side == 'n':
            plt.imshow(np.flip(np.array(dout_n_all).T),cmap=cmap,extent=[pout[-1],pout[0],vout[0],vout[-1]],aspect='auto')
            plt.ylabel('velocity (km/s)',fontsize=12)
            plt.xlabel('period (s)',fontsize=12)
            plt.xticks(np.linspace(pmin,pmax,5),fontsize=12)
            plt.yticks(np.linspace(vmin,vmax,5),fontsize=12)
            plt.clim(clim)
            ax=plt.colorbar()
            ax.set_label('normalized energy (%s)'%(energy_type.replace('_',' ')))
            plt.title('negative lag: '+energy_type.replace('_',' '),fontsize=13)
        elif side == 'p':
            plt.imshow(np.flip(np.array(dout_p_all).T),cmap=cmap,extent=[pout[-1],pout[0],vout[0],vout[-1]],aspect='auto')
            plt.ylabel('velocity (km/s)',fontsize=12)
            plt.xlabel('period (s)',fontsize=12)
            plt.xticks(np.linspace(pmin,pmax,5),fontsize=12)
            plt.yticks(np.linspace(vmin,vmax,5),fontsize=12)
            plt.clim(clim)
            ax=plt.colorbar()
            ax.set_label('normalized energy (%s)'%(energy_type.replace('_',' ')))
            plt.title('positive lag: '+energy_type.replace('_',' '),fontsize=13)
        #
        plt.show()

    if side=='a':
        dout=np.squeeze(np.array([dout_n_all,dout_p_all],dtype=np.float64))
    elif side == 'p':
        dout=np.squeeze(np.array(dout_p_all,dtype=np.float64))
    elif side == 'n':
        dout=np.squeeze(np.array(dout_n_all,dtype=np.float64))
    return dout,vout,pout
# function to extract the dispersion from the image
# modified from NoisePy.
def extract_dispersion_curve(amp,vel):
    '''
    this function takes the dispersion image as input, tracks the global maxinum on
    the spectrum amplitude

    PARAMETERS:
    ----------------
    amp: 2D amplitude matrix of the wavelet spectrum
    vel:  vel vector of the 2D matrix
    RETURNS:
    ----------------
    gv:   group velocity vector at each frequency
    '''
    nper = amp.shape[0]
    gv   = np.zeros(nper,dtype=np.float32)
    dv = vel[1]-vel[0]

    # find global maximum
    for ii in range(nper):
        maxvalue = np.max(amp[ii],axis=0)
        indx = list(amp[ii]).index(maxvalue)
        gv[ii] = vel[indx]

    return gv



import obspy,scipy,time,pycwt,pickle,os
import numpy as np
from obspy.signal.invsim import cosine_taper
from scipy.fftpack import fft,ifft,next_fast_len
from obspy.signal.filter import bandpass
from obspy import UTCDateTime
from seisgo.types import DvvData
from seisgo import utils,helpers
import matplotlib.pyplot as plt
from multiprocessing import Pool
import pyasdf
from scipy.signal import correlate,correlation_lags
from scipy.stats import linregress

'''
The dvv measuring functions are from Congcong Yuan at Harvard, originally written by Chengxin Jiang and Marine Denolle.
Xiaotao Yang adapted them for SeisGo and added wrapper functions.

Note by Congcong: several utility functions are modified based on https://github.com/tclements/noise
'''

########################################################
################ WRAPPER FUNCTIONS ##################
########################################################
def get_dvv(corrdata,freq,win,ref=None,stack_method='linear',offset=1.0,resolution=None,
            vmin=1.0,normalize=True,whiten='no',whiten_smooth=20, whiten_pad=100,
            method='wts',dvmax=0.05,subfreq=True,plot=False,figsize=(8,8),
            savefig=False,figdir='.',figname=None,figformat='png',save=False,outdir='.',outfile=None,
            format=None,nproc=None,v=False):
    """
    Compute dvv with given corrdata object, with options to save dvvdata to file.

    =====PARAMETERS=======
    corrdata: CorrData object that stores the correlaiton result.
    freq: minimum and maximum frequencies for dvv measurements.
    win: window length for dvv measurements.
    ref: reference trace/stack. default is None (will get by stacking all in the data)
    dvmax: maximum dvv searching range. default: 0.05 (5%).
    vmin: minimum velocity for the main phase, only considered in cross-correlations between
        two stations. default 1.0 km/s.
    offset: offset from 0.0 seconds (for autocorr) and from the maximum arrival time of the main
        phase (for xcorr)
    resolution: in seconds, specifying the temporal resolution (resampling/substacking) before measuring
        dvv.
    stack_method: stacking method to get the reference trace (if not given) and the short-window substacks. default is 'linear'.
    normalize: Ture or False for data normalization in measuring dvv.
    whiten='no',whiten_smooth=20, whiten_pad=100: parameters for whitening the trace before measuring dv/v.
                whiten: default is 'no', could be 'phase_only' or 'rma'.
    method: dvv measuring method.
    subfreq: keep all frequencies in the dvv result. default is True. Otherwise, only get one dvv result.
    plot: Default is False. It determines whether plots the corrdata and the measuring time windows.
        Plotting for dvvdata is currently seperated as a dvvdata.plot() method.
    savefig: Default False. Save plot or not.
    figdir: directory to save the figure.
    figname: figure name, excluding directory.
    save: this flag is for the dvvdata result. if true, the result will be saved to an ASDF file.
        Othersie, it returns the dvvdata object. Default is False.
    outdir: this is the directory to save the dvvdata.
    outfile: specify the file name to save the dvvdata. the file format can be indicated with the
            extension or "format"
    format: data file format: "asdf" or "pickle". Default is None, which will be determined by file extension.
    v: verbose. Default False.
    """
    method_all=helpers.dvv_methods() #list of all available methods.
    if method.lower() == "ts":
        subfreq=False
    if method.lower() not in method_all:
        raise ValueError(method+" is not available yet. Please change to one of: "+str(method_all))
    if corrdata.side.lower() != "a":
        raise ValueError("only works for now on corrdata with both sides. here: corrdata.side="+corrdata.side)
    # load stacked and sub-stacked waveforms
    cdata=corrdata.copy()
    #demean and detrend
    datatemp=cdata.data.copy()
    cdata.data=utils.demean(utils.detrend(datatemp))
    if resolution is not None: cdata.stack(win_len=resolution,method=stack_method,overwrite=True)
    if ref is None: ref=cdata.stack(method=stack_method,overwrite=False)

    nwin=cdata.data.shape[0]

    # make conda window based on vmin
    if cdata.sta[0]!=cdata.sta[1]: #xcorr
        tmin = offset+cdata.dist/vmin
    else:
        tmin=offset

    twin = [tmin,tmin+win]
    if twin[1] > cdata.lag:
        raise ValueError('proposed window exceeds limit! reduce %d'%win)

    # ref and tvec
    tvec_all = np.arange(-cdata.lag,cdata.lag+0.5*cdata.dt,cdata.dt)
    zero_indx0 = np.where((tvec_all> -0.5*cdata.dt)&(tvec_all<0.5*cdata.dt))[0]
    zero_indx = zero_indx0[np.argmin(np.abs(tvec_all[zero_indx0]))]
    tvec_half=tvec_all[zero_indx:]
    # casual and acasual coda window
    pwin_indx = np.where((tvec_all>=np.min(twin))&(tvec_all<=np.max(twin)))[0]
    nwin_indx = np.where((tvec_all<=-np.min(twin))&(tvec_all>=-np.max(twin)))[0]
    pcor_cc = np.zeros(shape=(nwin),dtype=np.float32)
    ncor_cc = np.zeros(shape=(nwin),dtype=np.float32)
    pcur=np.zeros(shape=(nwin,zero_indx+1),dtype=np.float32)
    ncur=np.zeros(shape=(nwin,zero_indx+1),dtype=np.float32)
    pref=np.zeros(shape=(nwin,zero_indx+1),dtype=np.float32)
    nref=np.zeros(shape=(nwin,zero_indx+1),dtype=np.float32)
    # allocate matrix for cur and ref waveforms and corr coefficient
    cur  = cdata.data #cdata.data
    # load all current waveforms and get corr-coeff
    if normalize:
        ref /= np.max(np.abs(ref))
        # loop through each cur waveforms
        for ii in range(nwin):
            cur[ii] /= np.max(np.abs(cur[ii]))
    #do whitening if specified.
    if whiten != 'no':
        ref = utils.whiten(ref,cdata.dt,freq[0],freq[1],method=whiten,smooth=whiten_smooth,pad=whiten_pad)
        cur = utils.whiten(cur,cdata.dt,freq[0],freq[1],method=whiten,smooth=whiten_smooth,pad=whiten_pad)
    for ii in range(nwin):
        # get cc coeffient
        pcor_cc[ii] = np.corrcoef(ref[pwin_indx],cur[ii,pwin_indx])[0,1]
        ncor_cc[ii] = np.corrcoef(ref[nwin_indx],cur[ii,nwin_indx])[0,1]

        pcur[ii] = cur[ii,zero_indx:]
        ncur[ii] = np.flip(cur[ii,:zero_indx+1])
        pref[ii] = ref[zero_indx:]
        nref[ii] = np.flip(ref[:zero_indx+1])
    #######################
    ##### MONITORING #####
    dvv_pos,dvv_neg,freqall,maxcc_p,maxcc_n,error_p,error_n=[],[],[],[],[],[],[]
    if nproc is None or nproc<2: #regular loop
        # loop through each win again
        for ii in range(nwin):
            # casual and acasual lags for both ref and cur waveforms
            if v: print('working on window: '+str(UTCDateTime(cdata.time[ii]))+" ... "+str(ii+1)+"/"+str(nwin))
            if method.lower()=="wts":
                freq_p,dvv_p,dvv_error_p,cc_p,_ = wts_dvv(pref[ii],pcur[ii],tvec_half,twin,freq,\
                                                                         subfreq=subfreq,dvmax=dvmax)
                _,dvv_n,dvv_error_n,cc_n,_ = wts_dvv(nref[ii],ncur[ii],tvec_half,twin,freq,\
                                                                             subfreq=subfreq,dvmax=dvmax)
            elif method.lower()=="ts":
                dvv_p,dvv_error_p,cc_p,_ = ts_dvv(pref[ii],pcur[ii],tvec_half,twin,freq,\
                                                                         dvmax=dvmax)
                dvv_n,dvv_error_n,cc_n,_ = ts_dvv(nref[ii],ncur[ii],tvec_half,twin,freq,\
                                                                             dvmax=dvmax)
                #
                freq_p = freq
            else:
                raise ValueError(method+" is not available yet. Please change to one of: "+str(method_all))
            if ii==0: freqall=freq_p
            dvv_pos.append(dvv_p)
            dvv_neg.append(dvv_n)
            maxcc_p.append(cc_p)
            maxcc_n.append(cc_n)
            error_p.append(dvv_error_p)
            error_n.append(dvv_error_n)
    else: #use multiple processor for parallel preprocessing
        #parallel
        print('working on %d windows with %d workers.'%(nwin,nproc))
        p=Pool(int(nproc))
        if method.lower()=="wts":
            presults=p.starmap(wts_dvv,[(pref[ii],pcur[ii],tvec_half,\
                                        twin,freq,subfreq,dvmax) for ii in range(nwin)])
            nresults=p.starmap(wts_dvv,[(nref[ii],ncur[ii],tvec_half,\
                                        twin,freq,subfreq,dvmax) for ii in range(nwin)])
            for ii in range(nwin):
                if ii==0: freqall=presults[ii][0]
                dvv_pos.append(presults[ii][1])
                dvv_neg.append(nresults[ii][1])
                error_p.append(presults[ii][2])
                error_n.append(nresults[ii][2])
                maxcc_p.append(presults[ii][3])
                maxcc_n.append(nresults[ii][3])
        elif method.lower()=="ts":
            #set filter to True when using as a standalone
            presults=p.starmap(ts_dvv,[(pref[ii],pcur[ii],tvec_half,\
                                        twin,freq,dvmax) for ii in range(nwin)])
            nresults=p.starmap(ts_dvv,[(nref[ii],ncur[ii],tvec_half,\
                                        twin,freq,dvmax) for ii in range(nwin)])
            for ii in range(nwin):
                if ii==0: freqall=freq
                dvv_pos.append(presults[ii][0])
                dvv_neg.append(nresults[ii][0])
                error_p.append(presults[ii][1])
                error_n.append(nresults[ii][1])
                maxcc_p.append(presults[ii][2])
                maxcc_n.append(nresults[ii][2])
        else:
            raise ValueError(method+" is not available yet. Please change to one of: "+str(method_all))
        p.close()

    #
    del pcur,ncur,pref,nref
    maxcc_p=np.array(maxcc_p)
    maxcc_n=np.array(maxcc_n)
    error_p=np.array(error_p)
    error_n=np.array(error_n)
    #for now, if errors are negative, assign np.nan to dvv data.
    dvv_neg=np.array(dvv_neg)
    dvv_pos=np.array(dvv_pos)
    idx1=np.where((error_n<0))
    idx2=np.where((error_p<0))
    error_n[idx1]=np.nan
    error_p[idx2]=np.nan
    maxcc_n[idx1]=np.nan
    maxcc_p[idx2]=np.nan
    dvv_neg[idx1]=np.nan
    dvv_pos[idx2]=np.nan
    dvvdata=DvvData(cdata,subfreq=subfreq,freq=freqall,cc1=ncor_cc,cc2=pcor_cc,maxcc1=maxcc_n,maxcc2=maxcc_p,
                        method=method,stack_method=stack_method,error1=error_n,error2=error_p,
                        window=twin,normalize=normalize,data1=np.array(dvv_neg),data2=np.array(dvv_pos))
    if save:
        dvvdata.save(outdir=outdir,file=outfile,format=format)

    ######plotting
    if plot:
        disp_indx = np.where(np.abs(tvec_all)<=np.max(twin)+0.2*win)[0]
        tvec_disp=tvec_all[disp_indx]
        # tick inc for plotting
        if nwin>100:
            tick_inc = int(nwin/10)
        elif nwin>10:
            tick_inc = int(nwin/5)
        else:
            tick_inc = 2
        #filter corrdata:
        for i in range(nwin):
            if freq[1]==0.5/cdata.dt:
                cur[i,:]=bandpass(cur[i,:],freq[0],0.995*freq[1],df=1/cdata.dt,corners=4,zerophase=True)
            else:
                cur[i,:]=bandpass(cur[i,:],freq[0],freq[1],df=1/cdata.dt,corners=4,zerophase=True)
        ref = bandpass(ref,freq[0],freq[1],df=1/cdata.dt,corners=4,zerophase=True)

        fig=plt.figure(figsize=figsize, facecolor = 'white')
        ax0= fig.add_subplot(8,1,(1,4))
        # 2D waveform matrix
        ax0.matshow(cur[:,disp_indx],cmap='seismic',extent=[tvec_disp[0],tvec_disp[-1],nwin,0],
                    aspect='auto')
        ax0.plot([0,0],[0,nwin],'k--',linewidth=2)
        ax0.set_title('%s, dist:%5.2fkm, %4.2f-%4.2f Hz' % (cdata.id,cdata.dist,freq[0],freq[1]))
        ax0.set_xlabel('time [s]')
        ax0.set_ylabel('waveforms')
        ax0.set_yticks(np.arange(0,nwin,step=tick_inc))
        # shade the coda part
        ax0.fill(np.concatenate((tvec_all[nwin_indx],np.flip(tvec_all[nwin_indx],axis=0)),axis=0), \
            np.concatenate((np.ones(len(nwin_indx))*0,np.ones(len(nwin_indx))*nwin),axis=0),'c', alpha=0.3,linewidth=1)
        ax0.fill(np.concatenate((tvec_all[pwin_indx],np.flip(tvec_all[pwin_indx],axis=0)),axis=0), \
            np.concatenate((np.ones(len(pwin_indx))*0,np.ones(len(pwin_indx))*nwin),axis=0),'y', alpha=0.3)
        ax0.xaxis.set_ticks_position('bottom')
        # reference waveform
        ax1 = fig.add_subplot(8,1,(5,6))
        ax1.plot(tvec_disp,ref[disp_indx],'k-',linewidth=1)
        ax1.autoscale(enable=True, axis='x', tight=True)
        ax1.grid(True)
        ax1.legend(['reference'],loc='upper right')

        # the cross-correlation coefficient
        xticks=np.int16(np.linspace(0,nwin-1,6))
        xticklabel=[]
        for x in xticks:
            xticklabel.append(str(UTCDateTime(cdata.time[x]))[:10])
        ax2 = fig.add_subplot(8,1,(7,8))
        ax2.plot(cdata.time,ncor_cc,'yo-',markersize=2,linewidth=1)
        ax2.plot(cdata.time,pcor_cc,'co-',markersize=2,linewidth=1)
        ax2.set_xticks(cdata.time[xticks])
        ax2.set_xticklabels(xticklabel,fontsize=12)
        ax2.set_title('substacked with '+stack_method,fontsize=12)
        # ax2.set_xticks(timestamp[0:nwin:tick_inc])
        ax2.set_xlim([min(cdata.time),max(cdata.time)])
        ax2.set_ylabel('cc coeff')
        ax2.legend(['negative','positive'],loc='upper right')

        plt.tight_layout()

        ###################
        ##### SAVING ######
        if savefig:
            if not os.path.isdir(figdir):os.mkdir(figdir)
            if figname is None:
                figname = 'xcc_'+cdata.id+'_'+cc_comp+'.'+figformat
            plt.savefig(figdir+'/'+figname, format=figformat, dpi=300, facecolor = 'white')
            plt.close()
        else:
            plt.show()

    if not save: return dvvdata
#
def extract_dvvdata(sfile,pair=None,comp=['all'],format=None):
    """
    Extracts DvvData from a file.

    PARAMETERS:
    --------------------------
    dfile: data file containing DvvData saved through DvvData.save().
    pair: net1.sta1-net2.sta2 pair to extract, default is to extract all pairs.
    comp: cross-correlation component or a list of components to extract, default is all components.
    format: if the file format is known, specify here. otherwise, the program determines it from the file extension.
            You can specify as "asdf" or "pickle"

    RETURN:
    --------------------------
    dvvdict: a dictionary that contains all extracted DvvData objects, which each key as the station
                pair name. for each station pair, the dvvdata are saved as a list of DvvData objects.
    USAGE:
    --------------------------
    extract_dvvdata('temp.h5',comp='ZZ')
    """
    #check help or not at the very beginning

    # open data for read
    if isinstance(pair,str): pair=[pair]
    if isinstance(comp,str): comp=[comp]

    #
    if format is None: #automatically determine the format.
        fext=sfile[-2:]
        if fext.lower() == "h5":
            format="asdf"
        elif fext.lower() == "pk":
            format="pickle"
        else:
            raise ValueError("file extension ["+fext+"] is not supported. Specify the format if you know it.")

    dvvdict=dict()
    if format.lower() == "asdf":
        try:
            ds = pyasdf.ASDFDataSet(sfile,mpi=False,mode='r')
            # extract common variables
            spairs_all = ds.auxiliary_data.list()
        except Exception:
            raise IOError("exit! cannot open %s to read"%sfile)
    elif format.lower() == "pickle":
        datain=pickle.load(open(sfile,"rb"))
        spairs_all=list(datain.keys())

    if pair is None: pair=spairs_all

    for spair in list(set(pair) & set(spairs_all)):
        ttr = spair.split('_')
        snet,ssta = ttr[0].split('.')
        rnet,rsta = ttr[1].split('.')
        if format.lower() == "asdf":
            path_lists = ds.auxiliary_data[spair].list()
        elif format.lower() == "pickle":
            path_lists = list(datain[spair].keys())
        #
        dvvdict[spair]=dict()
        for ipath in path_lists:
            schan,rchan = ipath.split('_')
            cc_comp=schan[-1]+rchan[-1]
            if cc_comp in comp or comp == ['all'] or comp ==['ALL']:
                try:
                    if format.lower() == "asdf":
                        para=ds.auxiliary_data[spair][ipath].parameters
                    elif format.lower() == "pickle":
                        para=datain[spair][ipath]["parameters"]
                    #
                    dt,dist,dist_unit,azi,baz,slon,slat,sele,rlon,rlat,rele,\
                    window,stack_method,method,normalize,ttime,comp,freq,\
                    net,sta,chan,side,cc1,cc2,maxcc1,maxcc2,error1,error2=\
                        para['dt'],para['dist'],para['dist_unit'],para['azi'],para['baz'],\
                        para['lonS'],para['latS'],para['eleS'],para['lonR'],para['latR'],para['eleR'],\
                        para['window'],para['stack_method'],para['method'],para['normalize'],\
                        para['time'],para['comp'],para['freq'],\
                        para['net'],para['sta'],para['chan'],para['side'],para['cc1'],para['cc2'],\
                        para['maxcc1'],para['maxcc2'],para['error1'],para['error2']

                    ##special handling of time, in case time_mean is saved to reduce
                    #the attribute memory_size
                    if "time_mean" in list(para.keys()):
                        tmean=para["time_mean"]
                        ttime = np.float64(ttime) + tmean
                    if "subfreq" in list(para.keys()):
                        subfreq=para['subfreq']
                    else:
                        subfreq=True #to be compatible with old usage, only wts with subfreq True.
                    if format.lower() == "asdf":
                        datamatrix=ds.auxiliary_data[spair][ipath].data
                    elif format.lower() == "pickle":
                        datamatrix=datain[spair][ipath]["data"]

                    if side.lower() == 'a':
                        data1 = datamatrix[0].copy()
                        data2 = datamatrix[1].copy()
                    elif side.lower()=='n':
                        data1 = datamatrix.copy()
                        data2=None
                    elif side.lower()=='p':
                        data1=None
                        data2 = datamatrix.copy()

                except Exception:
                    print('continue! something wrong with %s %s'%(spair,ipath))
                    continue
                dvvdict[spair][cc_comp]=DvvData(net=[snet,rnet],sta=[ssta,rsta],loc=['',''],\
                                                chan=chan,lon=[slon,rlon],lat=[slat,rlat],
                                                ele=[sele,rele],cc_comp=comp,dt=dt,dist=dist,
                                                dist_unit=dist_unit,az=azi,baz=baz,time=ttime,normalize=normalize,
                                                freq=freq,cc1=cc1,cc2=cc2,maxcc1=maxcc1,maxcc2=maxcc2,
                                                error1=error1,error2=error2,window=window,subfreq=subfreq,
                                                stack_method=stack_method,method=method,
                                                data1=data1,data2=data2)
        return dvvdict
    else:
        raise ValueError(format+" is not supported yet.")

##########################################################################
################ MONITORING FUNCTIONS ADAPTED FOR SEISGO ##################
##########################################################################
def ts_dvv(ref, cur, t,twin, freq, dvmax=0.05, ndv=100, filter=True):
    """
    This function compares the Reference waveform to stretched/compressed current waveforms to get the relative seismic velocity variation (and associated error).
    It also computes the correlation coefficient between the Reference waveform and the current waveform.
    PARAMETERS:
    ----------------
    ref: Reference waveform (np.ndarray, size N)
    cur: Current waveform (np.ndarray, size N)
    freq: [min,max] frequency frequency of the data
    t: time vector for the data.
    twin: time window for the measurements of dv/v
    dvmax: absolute bound for the velocity variation; example: dv=0.03 for [-3,3]% of relative velocity change ('float'). Default=0.05.
    ndv: number of stretching coefficient between dvmin and dvmax, no need to be higher than 100  ('float'). Default is 100.
    filter: apply filter with the specified frequency range. Default is True.
    RETURNS:
    ----------------
    dv: Relative velocity change dv/v (in %)
    cc: correlation coefficient between the reference waveform and the best stretched/compressed current waveform
    cdp: correlation coefficient between the reference waveform and the initial current waveform
    error: Errors in the dv/v measurements based on Weaver et al (2011), On the precision of noise-correlation interferometry, Geophys. J. Int., 185(3)
    Note: The code first finds the best correlation coefficient between the Reference waveform and the stretched/compressed current waveform among the "ndv" values.
    A refined analysis is then performed around this value to obtain a more precise dv/v measurement .
    Originally by L. Viens 04/26/2018 (Viens et al., 2018 JGR)
    modified by Chengxin Jiang, Xiaotao Yang
    """

    # load common variables from dictionary
    dt   = t[1]-t[0]
    tmin = np.min(twin)
    tmax = np.max(twin)
    fmin = np.min(freq)
    fmax = np.max(freq)
    itvec = np.arange(int((tmin-t.min())/dt)+1, int((tmax-t.min())/dt)+1)
    tvec = t[itvec]

    # make useful one for measurements
    dvmin = -np.abs(dvmax)
    dvmax = np.abs(dvmax)
    Eps = 1+(np.linspace(dvmin, dvmax, ndv))
    cof = np.zeros(Eps.shape,dtype=np.float32)

    #apply filter if requested
    if filter: #this is important when the data was not filtered before calling ts_dvv().
        ref=bandpass(ref,freq[0],0.998*freq[1],df=1/dt,corners=4,zerophase=True)
        cur=bandpass(cur,freq[0],0.998*freq[1],df=1/dt,corners=4,zerophase=True)
    refwin_temp=ref[itvec]
    curwin_temp=cur[itvec]
    refwin=refwin_temp/np.max(np.abs(refwin_temp))
    curwin=curwin_temp/np.max(np.abs(curwin_temp))
    # Set of stretched/compressed current waveforms
    # plt.figure()
    # plt.plot(tvec,refwin,'r',label='ref')
    # plt.plot(tvec,curwin,'b',label='cur')
    # plt.legend()
    # plt.grid()
    # plt.show()

    # plt.figure()
    for ii in range(len(Eps)):
        nt = tvec*Eps[ii]
        s = np.interp(x=tvec, xp=nt, fp=curwin)
        tcheckmin=np.max([nt[0],tvec[0]]) #trim the zero paddings due to extrapolation.
        tcheckmax=np.min([nt[-1],tvec[-1]])
        tcheckidx=np.where((tvec>= tcheckmin) & (tvec<=tcheckmax))[0]
        waveform_ref = refwin[tcheckidx]
        waveform_cur = s[tcheckidx]
        # plt.plot(tvec[tcheckidx],ii+waveform_ref,'r',label='ref')
        # plt.plot(tvec[tcheckidx],ii+waveform_cur,'b')
        cof[ii] = np.corrcoef(waveform_ref, waveform_cur)[0, 1]
#         cof[ii] = np.sum(np.power(waveform_ref-waveform_cur,2))
    #
    # plt.grid()
    # plt.show()
    cdp = np.corrcoef(curwin, refwin)[0, 1] # correlation coefficient between the reference and initial current waveforms

    # find the maximum correlation coefficient
    imax = np.nanargmax(cof)
    if imax >= len(Eps)-2:
        imax = imax - 2
    if imax <= 2:
        imax = imax + 2
    #
    # plt.plot(Eps,cof,'o')
    # plt.show()
    # Proceed to the second step to get a more precise dv/v measurement
    dtfiner = np.linspace(Eps[imax-2], Eps[imax+2], ndv)
    ncof    = np.zeros(dtfiner.shape,dtype=np.float32)

    for ii in range(len(dtfiner)):
        nt = tvec*dtfiner[ii]
        s = np.interp(x=tvec, xp=nt, fp=curwin)
        tcheckmin=np.max([nt[0],tvec[0]])
        tcheckmax=np.min([nt[-1],tvec[-1]])
        tcheckidx=np.where((tvec>= tcheckmin) & (tvec<=tcheckmax))[0]
        waveform_ref = refwin[tcheckidx]
        waveform_cur = s[tcheckidx]
        ncof[ii] = np.corrcoef(waveform_ref, waveform_cur)[0, 1]
    cc = np.max(ncof) # Find maximum correlation coefficient of the refined  analysis
    dv = 100. * dtfiner[np.argmax(ncof)]-100 # Multiply by 100 to convert to percentage (Epsilon = -dt/t = dv/v)

    # Error computation based on Weaver et al (2011), On the precision of noise-correlation interferometry, Geophys. J. Int., 185(3)
    T = 1 / (fmax - fmin)
    X = cc
    wc = np.pi * (fmin + fmax)
    t1 = np.min([tmin, tmax])
    t2 = np.max([tmin, tmax])
    error = 100*(np.sqrt(1-X**2)/(2*X)*np.sqrt((6* np.sqrt(np.pi/2)*T)/(wc**2*(t2**3-t1**3))))

    return dv, error, cc, cdp

def wts_dvv(ref,cur,t,twin,freq,subfreq=True,dvmax=0.05,normalize=True,ndv=100,dj=1/12,s0=-1,J=-1,wvn='morlet'):
    """
    Apply stretching method to continuous wavelet transformation (CWT) of signals
    for all frequecies in an interest range

    Parameters
    --------------
    ref: The complete "Reference" time series (numpy.ndarray)
    cur: The complete "Current" time series (numpy.ndarray)
    t: time vector (one side)
    twin: time window to measure dv/v
    freq: frequence range for measuring.
    subfreq: a boolen variable to make measurements on all frequency range or not. Default True.
    dvmax: absolute bound for the velocity variation; example: dv=0.03 for [-3,3]% of relative velocity change (float).
            Default: 0.05 (5%).
    ndv: number of stretching coefficient between dvmin and dvmax, no need to be higher than 100  (Default)
    dj, s0, J, sig, wvn: common parameters used in 'wavelet.wct'. Defaults are dj=1/12,s0=-1,J=-1,wvn='morlet'
    normalize: normalize the wavelet spectrum or not. Default is True

    RETURNS:
    ------------------
    dvv: estimated dv/v
    err: error of dv/v estimation

    Written by Congcong Yuan (30 Jun, 2019)
    Adapted by Xiaotao Yang for SeisGo (July, 2021)
    """
    # common variables
    dt   = t[1]-t[0]
    tmin = np.min(twin)
    tmax = np.max(twin)
    fmin = np.min(freq)
    fmax = np.max(freq)
    itvec = np.arange(np.int((tmin-t.min())/dt)+1, np.int((tmax-t.min())/dt)+1)

    # apply cwt on two traces
    cwt1, sj, f, coi, _, _ = pycwt.cwt(cur, dt, dj, s0, J, wvn)
    cwt2, sj, f, coi, _, _ = pycwt.cwt(ref, dt, dj, s0, J, wvn)

    # zero out data outside frequency band
    if (fmax> np.max(f)) | (fmax <= fmin):
        raise ValueError('Abort: input frequency out of limits!')
    else:
        f_ind = np.where((f >= fmin) & (f <= fmax))[0]

    # convert wavelet domain back to time domain (~filtering)
    if not subfreq:

        # inverse cwt to time domain
        icwt1 = pycwt.icwt(cwt1[f_ind], sj[f_ind], dt, dj, wvn)
        icwt2 = pycwt.icwt(cwt2[f_ind], sj[f_ind], dt, dj, wvn)

        # assume all time window is used
        wcwt1, wcwt2 = np.real(icwt1), np.real(icwt2)

        # Normalizes both signals, if appropriate.
        if normalize:
            ncwt1 = (wcwt1 - wcwt1.mean()) / wcwt1.std()
            ncwt2 = (wcwt2 - wcwt2.mean()) / wcwt2.std()
        else:
            ncwt1 = wcwt1
            ncwt2 = wcwt2

        # run stretching
        dvv, err, cc, cdp = ts_dvv(ncwt2, ncwt1, t,twin,freq,dvmax=dvmax, ndv=ndv,filter=False)
        return freq, dvv, err, cc, cdp

    # directly take advantage of the real-valued parts of wavelet transforms
    else:
        # extract real values of cwt
        rcwt1, rcwt2 = np.real(cwt1), np.real(cwt2)
        # initialize variable
        nfreq=len(f_ind)
        dvv, cc, cdp, err = np.zeros(nfreq,dtype=np.float32), np.zeros(nfreq,dtype=np.float32),\
            np.zeros(nfreq,dtype=np.float32),np.zeros(nfreq,dtype=np.float32)
        # loop through each freq
        for ii, ifreq in enumerate(f_ind):

            # prepare windowed data
            wcwt1, wcwt2 = rcwt1[ifreq], rcwt2[ifreq]

            # Normalizes both signals, if appropriate.
            if normalize:
                ncwt1 = (wcwt1 - wcwt1.mean()) / wcwt1.std()
                ncwt2 = (wcwt2 - wcwt2.mean()) / wcwt2.std()
            else:
                ncwt1 = wcwt1
                ncwt2 = wcwt2
            #update the frequency to be passed to ts_dvv(). the frequency range is used to compute the errors.
            if ii >0:
                if ii < len(f_ind) - 1:
                    newfreq=[np.mean([f[ifreq],f[f_ind[ii-1]]]),np.mean([f[ifreq],f[f_ind[ii+1]]])]
                else:
                    df=np.abs(f[f_ind[ii]] - f[f_ind[ii-1]])
                    newfreq=[np.mean([f[ifreq],f[f_ind[ii-1]]]),f[ifreq]+0.5*df]
            else:
                df=np.abs(f[f_ind[ii+1]] - f[f_ind[ii]])
                newfreq=[f[ifreq]-0.5*df,np.mean([f[ifreq],f[f_ind[ii+1]]])]
            # run stretching
            dv, error, c1, c2 = ts_dvv(ncwt2, ncwt1, t,twin,newfreq, dvmax=dvmax, ndv=ndv,filter=False)
            dvv[ii], err[ii], cc[ii], cdp[ii]=dv, error,c1, c2

        return f[f_ind], dvv, err, cc, cdp


#
def xc_dvv(ref, cur, t,twin, freq, filter=True,plot=False):
    """
    This function compares the waveform differences through moving cross-correlations. It might suffer from
    cycle skipping.
    PARAMETERS:
    ----------------
    ref: Reference waveform (np.ndarray, size N)
    cur: Current waveform (np.ndarray, size N)
    freq: [min,max] frequency frequency of the data
    t: time vector for the data.
    twin: time window for the measurements of dv/v
    filter: apply filter with the specified frequency range. Default is True.
    RETURNS:
    ----------------
    dv: Relative velocity change dv/v (in %)
    cc: is the R-squared from the linear fit.
    cdp: correlation coefficient between the reference waveform and the initial current waveform
    error: Errors in the dv/v measurements based on linear regression.
    """

    # load common variables from dictionary
    dt   = t[1]-t[0]
    tmin = np.min(twin)
    tmax = np.max(twin)
    fmin = np.min(freq)
    fmax = np.max(freq)
    itvec = np.arange(int((tmin-t.min())/dt)+1, int((tmax-t.min())/dt)+1)
    tvec = t[itvec]

    #apply filter if requested
    if filter: #this is important when the data was not filtered before calling ts_dvv().
        ref=bandpass(ref,freq[0],0.998*freq[1],df=1/dt,corners=4,zerophase=True)
        cur=bandpass(cur,freq[0],0.998*freq[1],df=1/dt,corners=4,zerophase=True)
    refwin_temp=ref[itvec]
    curwin_temp=cur[itvec]
    refwin=refwin_temp/np.max(np.abs(refwin_temp))
    curwin=curwin_temp/np.max(np.abs(curwin_temp))
    cdp = np.corrcoef(curwin, refwin)[0, 1] # correlation coefficient between the reference and initial current waveforms

    #
    winlen=int(2/freq[0]/dt)
    step=int(winlen/2)
    sliceref,slicen,sliceidx=utils.sliding_window(refwin,winlen,ss=step,getindex=True)
    slicecur,slicen,sliceidx=utils.sliding_window(curwin,winlen,ss=step,getindex=True)
    lags = dt*correlation_lags(winlen, winlen)
    dtarray=[]
    tarray=[]
    cc_all=[]
    for i in range(slicen-1):
        xc = correlate(slicecur[i],sliceref[i])
        cc_all.append(np.max(xc))

        xcmax=np.nanargmax(xc)
        xcmaxlag=lags[xcmax]
        dtarray.append(xcmaxlag)
        tarray.append(tvec[sliceidx[i]])

    #
    tarray=np.array(tarray)
    dtarray=np.array(dtarray)
    res=linregress(tarray,dtarray)
    if plot:
        plt.figure()
        plt.plot(tarray,dtarray,'o')
        plt.plot(tarray, res.intercept + res.slope*tarray, 'r', label='fitted line')
        plt.show()

    dv=-100*res.slope
    error=100*res.stderr
    cc=np.power(res.rvalue,2)

    return dv, error, cc, cdp



#!/usr/bin/env python
# coding: utf-8
import os,sys,time
import numpy as np
import pandas as pd
from seisgo import noise,utils
from multiprocessing import Pool
import pygmt as gmt
from seisgo.anisotropy import do_BANX
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning) 
# ## BANX wrapper with control parameters
def BANX_wrapper(stationdict_all, reference_site, datadir, outdir_root, receiver_box):
    """
    Wrapper function for do_BANX function in anisotropy module.
    This function is called by the main script mainly for parallelization purpose.
    """
    XCorrComp = 'ZZ'
    # Subarray parameters:
    Min_Stations = 10   

    # SNR:
    Min_SNR = 10

    # Scaling factors
    Min_Radius_scaling = 1
    Max_Radius_scaling = 1.25
    Min_Distance_scaling = 3
    Signal_Extent_Scaling = 2.5 # used to set the signal window when computing the signal to noise ratio. 
    # The window would be [predicted arrival time - Signal_Extent_Scaling*Max_Period, predicted arrival time + Signal_Extent_Scaling*Max_Period] 

    Vel_Signal_Window = 3.2
    # Beamforming space:
    Max_Slowness = 0.5   # [s/km] Maximum slowness in the beamforming
    Slowness_Step = 0.005  #[s/km] Slowness interval

    # Beamforming limit:
    Vel_Reference = 3.5  # [km/s]
    Vel_Perturbation = 0.4 # [Percentage fraction] 0.5 = %50                                                                                              

    Taper_Length_Scaling= 7 # Taper length scaling factor. The taper length is Taper_Length_Scaling*Max_Period.

    AZIBIN_STEP = 6 # azimuthal bin step size in degrees used in the QC step after beamforming of all sources.
    # QC baz coverage
    Min_BAZ_measurements = 1 #minimum number of measurements in each azimuthal bin. Should be >=3. Use 1 for testing here.
    Min_Good_BAZBIN = 5 #minimum number of good bins with >= Min_BAZ_measurements in each azimuthal bin. Should be >=5 (recommended).

    Min_Beam_Sharpness = 10 #minimum beam sharpness to be considered as a good measurement.

    MinTime = 0.0 #start time of the xcorr data.

    Sampling_Rate_Target = 5 #target sampling rate. Needs to be integer times the data sampling rate to avoid resampling error.
    #The data will be resampled to Sampling_Rate_Target (samples per second).

    Period_Band = [15,30]

    DoubleSided = False

    ####################################################
    #### plotting controls ##############################
    ####################################################
    show_fig = False #figures will be plotted by not shown.
    # plot moveout of good traces
    plot_moveout = True
    moveout_scaling = 4

    # plot cluster map and the source
    plot_clustermap = True #not plot if on HPC cluster. some display issue may happend if plotting.
    map_engine = 'cartopy'
    #slowness image
    plot_beampower =True

    # plot phase velocity of the reference station
    plot_station_result = True

    ########################################################
    #### Calling do_BANX function in anisotropy module. ####
    ########################################################
    Beam_Local, anisotropy=do_BANX(stationdict_all, reference_site, Period_Band, Vel_Reference,datadir,
                                    outdir_root,sampling_rate=Sampling_Rate_Target,min_stations=Min_Stations, 
                                    min_snr=Min_SNR, min_radius_scaling=Min_Radius_scaling,
                                    max_radius_scaling=Max_Radius_scaling, min_distance_scaling=Min_Distance_scaling, 
                                    signal_window_velocity=Vel_Signal_Window,
                                    signal_extent_scaling=Signal_Extent_Scaling,max_slowness=Max_Slowness,slowness_step=Slowness_Step,
                                    velocity_perturbation=Vel_Perturbation, trace_start_time=MinTime,taper_length_scaling=Taper_Length_Scaling,
                                    azimuth_step=AZIBIN_STEP,min_baz_measurements=Min_BAZ_measurements,min_good_bazbin=Min_Good_BAZBIN,
                                    min_beam_sharpness=Min_Beam_Sharpness,doublesided=DoubleSided, cc_comp =XCorrComp,
                                    receiver_box=receiver_box,show_fig=show_fig,plot_moveout=plot_moveout, moveout_scaling=moveout_scaling,
                                    plot_clustermap=plot_clustermap, map_engine = map_engine, plot_beampower=plot_beampower,
                                    plot_station_result=plot_station_result,verbose=True)
    
    return Beam_Local, anisotropy

## Main script
# This is the main script for BANX processing.
def main():
    """
    Main script for BANX processing.
    """
    # Read number of processors from command line.
    narg = len(sys.argv)
    if narg == 1:
        nproc=1
    else:
        nproc=int(sys.argv[1])
    #
    if nproc > 1:
        # Create a pool of workers
        pool = Pool(processes=nproc)
    """
    Most of the time, users only need to change the following parameters:
    """
    # set root directory
    rootdir='.'
    datadir=os.path.join(rootdir,'data_craton/PAIRS_AVERAGEDSIDES_stack_robust_egf') #'data_craton/PAIRS_TWOSIDES_stack_robust')
    outdir_root=os.path.join(rootdir,'BANX_results')
    if not os.path.isdir(outdir_root):os.makedirs(outdir_root, exist_ok=True)
    ReceiverBox_lat=[36,39.5]
    ReceiverBox_lon=[-92,-84]
    stationinfo_file='station_info.csv'
    use_stationfile=True
    plot_station_map = True
    ######
    ##########################################################
    ####### End of user parameters. ############################
    ##########################################################

    # ## Extract the netsta list and their coordinates from the xcorr data
    # The coordinates for each net.sta are stored in dictionaries.
    # load data
    if use_stationfile:
        station_df=pd.read_csv(stationinfo_file)
        coord_all=dict()
        for i in range(len(station_df)):
            coord_all[station_df['net.sta'].iloc[i]] = [station_df['lat'].iloc[i],station_df['lon'].iloc[i],station_df['ele'].iloc[i]]
        # remove duplicates
        netsta_all=list(coord_all.keys()) #sorted(set(netsta_all))
        print('Read %d station from %s'%(len(netsta_all),stationinfo_file))
    else:
        sourcelist=utils.get_filelist(datadir)
        t1 = time.time()
        # netsta_all=[]
        coord_all=dict()
        if nproc < 2:
            for src in sourcelist:
                # srcdir=os.path.join(datadir,src)
                ccfiles=utils.get_filelist(src,'h5',pattern='P_stack')
                _,_,coord=noise.get_stationpairs(ccfiles,getcoord=True,verbose=True)
                # netsta_all.extend(netsta)
                coord_all = coord_all | coord
        else:
            #parallelization
            print('Using %d processes to process %d source files'%(nproc,len(sourcelist)))
            results=pool.starmap(noise.get_stationpairs, [(utils.get_filelist(src,'h5','P_stack'),
                                                        False,False,True) for src in sourcelist])
            # If running interactively, change the above line to: 
            # results = pool.startmap(noise.get_stationpairs, [(src,True) for src in sourcelist])
            # unpack results. Needed when running interactively. Otherwise, the results are not unpacked and have been saved to files.
            _, _, coord_all = zip(*results)
            # netsta_all = [item for sublist in netsta_all for item in sublist]
            coord_all = {k: v for d in coord_all for k, v in d.items()}
        #
        # remove duplicates
        netsta_all=list(coord_all.keys()) #sorted(set(netsta_all))
        print('Extracted %d net.sta from %d source files in %.2f seconds.'%(len(netsta_all),len(sourcelist),time.time()-t1))
        
        stationfile = os.path.join(rootdir,stationinfo_file)
        fout = open(stationfile,'w')
        fout.write('net.sta,lat,lon,ele\n')
        for i in range(len(netsta_all)):
            coord0 = coord_all[netsta_all[i]]
            fout.write('%s,%f,%f,%f\n'%(netsta_all[i],coord0[0],coord0[1],coord0[2]))
        fout.close()
        print('Station information saved to %s'%stationfile)

    # ## Subset the station list for the receiver box region
    #set receiver region box
    #this is usually a smaller region than the entire dataset. 
    # Stations within this box region are used as receivers while all
    # stations may be used as the sources.
    # set receiver box for do_BANX function.
    ReceiverBox = [ReceiverBox_lon[0],ReceiverBox_lon[1],ReceiverBox_lat[0],ReceiverBox_lat[1]]

    ReceiverList_Sites=[] #net.sta strings.
    ReceiverList_Coord=[] #lat, lon

    SourceList_Sites=[]
    SourceList_Coord=[]
    for i in range(len(netsta_all)):
        #Master site
        coord0 = coord_all[netsta_all[i]][:2] #coordinates: lat, lon in order.
        SourceList_Sites.append(netsta_all[i])
        SourceList_Coord.append(coord0)
        if coord0[0] >= ReceiverBox_lat[0] and coord0[0] <= ReceiverBox_lat[1] and \
                    coord0[1] >= ReceiverBox_lon[0] and coord0[1] <= ReceiverBox_lon[1]:
            ReceiverList_Sites.append(netsta_all[i])
            ReceiverList_Coord.append(coord0)
    #
    """
    Plot the station map.
    """
    if plot_station_map:
        #plot station map.
        source_coord_array=np.array(SourceList_Coord)
        marker_style="i0.17c"
        map_style="plain"
        projection="M3.i"
        frame="af"
        title="station map"
        GMT_FONT_TITLE="14p,Helvetica-Bold"
        lon_all,lat_all=source_coord_array[:,1],source_coord_array[:,0]

        region="%6.2f/%6.2f/%5.2f/%5.2f"%(np.min(lon_all),np.max(lon_all),np.min(lat_all),np.max(lat_all))
        fig = gmt.Figure()
        gmt.config(MAP_FRAME_TYPE=map_style, FONT_TITLE=GMT_FONT_TITLE)
        fig.coast(region=region, resolution="f",projection=projection, 
                water="0/180/255",frame=frame,land="240",
                borders=["1/1p,black", "2/0.5p,100"])
        fig.basemap(frame='+t'+title+'')
        fig.plot(
            x=lon_all,
            y=lat_all,
            style=marker_style,
            pen="0.5p,red",
        )
        #plot receiver box
        lon_box=[ReceiverBox_lon[0],ReceiverBox_lon[1],ReceiverBox_lon[1],ReceiverBox_lon[0],ReceiverBox_lon[0]]
        lat_box=[ReceiverBox_lat[0],ReceiverBox_lat[0],ReceiverBox_lat[1],ReceiverBox_lat[1],ReceiverBox_lat[0]]
        fig.plot(
            x=lon_box,
            y=lat_box,
            pen="1p,blue",
        )
        fig.savefig(os.path.join(rootdir,'station_map.pdf'))
        gmt.set_display('none')
        fig.show()
    #

    """"
    Start the main loop
    """
    #######################################
    #### Loop over the reference sites ####
    #######################################
    if nproc <2:
        # Beam_Local_all, anisotropy_all = [],[]
        for i in range(len(ReceiverList_Sites)):
            #Master site
            Ref_Site = ReceiverList_Sites[i]
            print('Processing reference site %s --- %d/%d'%(Ref_Site,i+1,len(ReceiverList_Sites)))
            
            # Beam_Local, anisotropy=BANX_wrapper(coord_all,Ref_Site, datadir, outdir_root, ReceiverBox)
            _,_=BANX_wrapper(coord_all,Ref_Site, datadir, outdir_root, ReceiverBox)
            #end here for debug/test.
            # Beam_Local_all.append(Beam_Local)
            # anisotropy_all.append(anisotropy)
            # break
        #
    else:
        #parallelization
        print('Using %d processes to process %d receiver sites'%(nproc,len(ReceiverList_Sites)))
        ############
        
        pool.starmap(BANX_wrapper, [(coord_all,Ref_Site, datadir, outdir_root, ReceiverBox) for Ref_Site in ReceiverList_Sites])
        # If running interactively, change the above line to: 
        # results = pool.startmap(BANX_wrapper, [(coord_all,Ref_Site, datadir, outdir_root, ReceiverBox) for Ref_Site in ReceiverList_Sites])
        pool.close()

        # unpack results. Needed when running interactively. Otherwise, the results are not unpacked and have been saved to files.
        
        # Beam_Local_all, anisotropy_all = zip(*results)
##end of main

########### 
if __name__ == "__main__":
    main()
    






#!/usr/bin/env python
# coding: utf-8
import os,sys,glob
from seisgo.noise import compute_fft,correlate
from seisgo import utils,downloaders

rootdir='.'
respdir='.'
sacfiles = sorted(glob.glob(os.path.join(rootdir,'*.SAC')))
rm_resp='RESP'
#for removing responses.
freqmin=0.01
freqmax=100

tr,inv=downloaders.read_data(sacfiles,rm_resp=rm_resp,freqmin=freqmin,freqmax=freqmax,stainv=True)
tr1,tr2=tr;inv1,inv2=inv

#trimming is needed for this data set, which there is one sample difference in the starting time.
cstart=max([tr1.stats.starttime,tr2.stats.starttime])
cend=min([tr1.stats.endtime,tr2.stats.endtime])
tr1.trim(starttime=cstart,endtime=cend,nearest_sample=True)
tr2.trim(starttime=cstart,endtime=cend,nearest_sample=True)

print('cross-correlation ...')
cc_len    = 3600                                                            # basic unit of data length for fft (sec)
cc_step      = 900                                                             # overlapping between each cc_len (sec)
maxlag         = 100                                                        # lags of cross-correlation to save (sec)
freq_norm='rma'
time_norm='no'

#for whitening
freqmin=0.02
freqmax=2
#get FFT, #do correlation
fftdata1=compute_fft(tr1,cc_len,cc_step,stainv=inv1,
                        freq_norm=freq_norm,freqmin=freqmin,freqmax=freqmax,
                     time_norm=time_norm,smooth=500)
fftdata2=compute_fft(tr2,cc_len,cc_step,stainv=inv2,
                       freq_norm=freq_norm,freqmin=freqmin,freqmax=freqmax,
                     time_norm=time_norm,smooth=500)
corrdata=correlate(fftdata1,fftdata2,maxlag,substack=True)

#plot xcorr result
freqs=[[0.05,0.1],[0.07,0.1],[0.1,0.5],[0.1,1],[0.5,1],[1,2]]
for i in range(len(freqs)):
    corrdata.plot(freqmin=freqs[i][0],freqmax=freqs[i][1],lag=50,stack_method='robust',save=True)

corrdata.to_asdf('2020.087_xcorr.h5')

corrdata.to_sac()
